import socket
import ssl
import sys
import task_information
import time
import json
import base64
import threading

memory_lock = threading.RLock()

_version = "1.5"
def get_version():
    return _version

# --- not recommended to use outside this file --- #
default_server_socket: ssl.SSLSocket = None

# this function will initialize a server_socket.
# After calling this function you don't need to tell the other functions the server_socket anymore. But you still can.
def initialize_server_socket(server_socket: ssl.SSLSocket):
    global default_server_socket
    default_server_socket = server_socket

# --- not recommended to get called outside this file --- #
def recv_msg_in_bytes(server_socket: ssl.SSLSocket = None) -> bytes:
    if server_socket == None: server_socket = default_server_socket

    number_of_bytes = str(server_socket.recv(1024), 'utf8') # the server will send the amount of bytes he will send next

    number_of_bytes = int(number_of_bytes)

    # now we calculate that we recv number_of_bytes in a potency of 2
    potency = 1
    while(potency < number_of_bytes):
        potency *= 2
    number_of_bytes = potency # reassigning number_of_bytes with a potency of 2

    msg = server_socket.recv(number_of_bytes) # now we recv the actual msg from the server
    return msg

# --- not recommended to get called outside this file --- #
def recv_msg(server_socket: ssl.SSLSocket = None) -> str:
    if server_socket == None: server_socket = default_server_socket

    return str(recv_msg_in_bytes(server_socket=server_socket), 'utf8')

# --- not recommended to get called outside this file --- #
def recv_msg_as_list(server_socket: ssl.SSLSocket = None) -> list:
    if server_socket == None: server_socket = default_server_socket

    return list(recv_msg_in_bytes(server_socket=server_socket))

# --- not recommended to get called outside this file --- #
def recv_msg_as_json(server_socket: ssl.SSLSocket = None) -> dict:
    if server_socket == None: server_socket = default_server_socket

    return json.loads(recv_msg(server_socket=server_socket))

# --- not recommended to get called outside this file --- #
def send_msg_in_bytes(msg: bytes, server_socket: ssl.SSLSocket = None):
    if server_socket == None: server_socket = default_server_socket
    
    server_socket.send(bytes(str(len(msg)), 'utf8')) # sending the amount of bytes the server needs to recv
    server_socket.send(msg) # sending the actual msg to the server

# --- not recommended to get called outside this file --- #
def send_msg(msg: str, server_socket: ssl.SSLSocket = None):
    if server_socket == None: server_socket = default_server_socket

    b_msg = bytes(msg, 'utf8')
    send_msg_in_bytes(server_socket=server_socket, msg=b_msg)

# --- not recommended to get called outside this file --- #
def send_msg_as_list(msg: list, server_socket: ssl.SSLSocket = None):
    if server_socket == None: server_socket = default_server_socket

    l_msg = bytes(msg)
    send_msg_in_bytes(server_socket=server_socket, msg=l_msg)

# --- not recommended to get called outside this file --- #
def send_msg_as_json(msg: dict, indent = None, sort_keys=False, server_socket: ssl.SSLSocket = None):
    if server_socket == None: server_socket = default_server_socket

    json_msg = json.dumps(msg, indent=indent, sort_keys=sort_keys)
    send_msg(msg=json_msg, server_socket=server_socket)

# --- not recommended to get called outside this file --- #
def send_task(task: list, server_socket: ssl.SSLSocket = None):
    if server_socket == None: server_socket = default_server_socket

    json_dict = {"t": task[0]}
    if len(task) > 1:
        json_dict.update({"tp": task[1:]})

    send_msg_as_json(server_socket=server_socket, msg=json_dict)

# the server will not return "result" here. Instead of that he will return "valid" as key and a boolean as value
def start_task_api_version_check(server_socket: ssl.SSLSocket = None) -> dict:
    if server_socket == None: server_socket = default_server_socket

    json_dict = {"version": get_version()}

    send_msg_as_json(msg=json_dict, server_socket=server_socket)
    ret_val = recv_msg_as_json(server_socket=server_socket)
    return ret_val

# this function will return an integer, which tells you the outcome of the task:
# 0    successfull. You are now logged in.
# 1    Something went wrong. Probably the username or password is wrong.
# 2    Already logged in as another user.
def start_task_login(username: str, password: str, task_parameters: list = task_information.login.default_task_parameters, server_socket: ssl.SSLSocket = None) -> dict:
    if server_socket == None: server_socket = default_server_socket

    # starting the task
    task_list: list = [task_information.login.task_number] + task_parameters
    send_task(task=task_list, server_socket=server_socket)

    b64pwd: str = str(base64.b64encode(bytes(password, 'utf8')), 'utf8')
    information_dict = {"usr": username, "b64pwd": b64pwd}
    send_msg_as_json(msg=information_dict, server_socket=server_socket)

    # return the outcome of the task
    task_result: dict = recv_msg_as_json(server_socket=server_socket)
    return task_result
    
# this function will return an integer, which tells you the outcome of the task:
# 0    successfull. You are now logged in.
# 1    Something went wrong.
# 2    Something went wrong. The username is probably already given or not usable for some reason.
# 3    Password is too short. You need a minimum of 8 characters
def start_task_register(username: str, password: str, task_parameters: list = task_information.register.default_task_parameters, server_socket: ssl.SSLSocket = None) -> dict:
    if server_socket == None: server_socket = default_server_socket

    # starting the task
    task_list: list = [task_information.register.task_number] + task_parameters
    send_task(task=task_list, server_socket=server_socket)

    b64pwd: str = str(base64.b64encode(bytes(password, 'utf8')), 'utf8')
    information_dict = {"usr": username, "b64pwd": b64pwd}
    send_msg_as_json(msg=information_dict, server_socket=server_socket)

    # return the outcome of the task
    task_result: dict = recv_msg_as_json(server_socket=server_socket)
    return task_result

# this function will return an integer, which tells you the outcome of the task:
# 0    successfull. You are no longer logged in.
def start_task_logout(task_parameters: list = task_information.logout.default_task_parameters, server_socket: ssl.SSLSocket = None) -> dict:
    if server_socket == None: server_socket = default_server_socket

    # starting the task
    task_list: list = [task_information.logout.task_number] + task_parameters
    send_task(task=task_list, server_socket=server_socket)

    # return the outcome of the task
    task_result: dict = recv_msg_as_json(server_socket=server_socket)
    return task_result

# this function will return a string. The first index of the string will give you information about the outcome of the task, formatted as a binary number. So you can't just convert to int:
# 0     Everything was successfull
# 1     Undefined Error
# 2     Client isn't logged in
# 3     Label invalid
# The rest of the string can be some additional information
# Please be carefull to use the parameters correctly. Its all about repeat.
# If repeat is true, you need to give information for the start_point, stop_point and frequence, but not for the time_point. The start_point and stop_point can be left out. If so, it will start immediately and never stop. The frequence has to be given.
# If repeat is false, you don't need to give information for the start_point, stop_point and frequence, but for the time_point.
def start_task_memory(label: str = None, msg: str = None, receiver: str = "self", repeat: bool = False, start_point: float = None, stop_point: float = None, frequence: float = None, time_point = None,
                    task_parameters: list = task_information.memory.default_task_parameters, server_socket: ssl.SSLSocket = None) -> dict:
    global memory_lock
    with memory_lock:
        if server_socket == None: server_socket = default_server_socket

        # starting the task
        task_list: list = [task_information.memory.task_number] + task_parameters
        send_task(server_socket=server_socket, task=task_list)

        information_dict = {}
        # evaluate the task_parameters
        if task_parameters[0] == 0:
            if task_parameters[1] == 0:
                pass # nothing to do there
            elif task_parameters[1] == 1:
                if task_parameters[2] == 0: # Not implemented yet
                    information_dict.update({"memory_label": label})
                    send_msg_as_json(msg=information_dict, server_socket=server_socket)
                elif task_parameters[2] == 1:
                    information_dict.update({"memory_label": label})
                    send_msg_as_json(msg=information_dict, server_socket=server_socket)
        elif task_parameters[0] == 1: # We want to send a new memory task to the server
            # sending all information in the right order
            information_dict.update({"memory_label": label})
            information_dict.update({"memory_msg": msg})
            information_dict.update({"memory_receiver": receiver})
            information_dict.update({"memory_repeat": repeat})

            # We have to send different information based on repeat
            if repeat:
                if start_point == None: start_point = time.time() # giving start_point a value if it doesn't already have one
                if stop_point == None: stop_point: float = 0 # giving stop_point a value if it doesn't already have one
                information_dict.update({"memory_start_point": start_point})
                information_dict.update({"memory_stop_point": stop_point})
                information_dict.update({"memory_frequence": frequence})
            else:
                information_dict.update({"memory_time_point": time_point})
            send_msg_as_json(msg=information_dict, server_socket=server_socket)
        elif task_parameters[0] == 2:
            information_dict.update({"memory_label": label})
            send_msg_as_json(msg=information_dict, server_socket=server_socket)
        elif task_parameters[0] == 3:
            information_dict.update({"label": label})
            send_msg_as_json(msg=information_dict, server_socket=server_socket)
        
        # return the outcome of the task
        task_result: dict = recv_msg_as_json(server_socket=server_socket)
        return task_result

# the will return the outcome of the operation in json format (outcome is "result")
# 0     Everything was successfull
# 1     Unexpected Error (The token is probably still there)
# 2     Token invalid
# 3     Wrong format. Error with JSON.
# 4     User not logged in/already logged in
def start_task_login_token(username: str = None, token: str = None, old_token: str = None, task_parameters: list = task_information.login_token.default_task_parameters, server_socket: ssl.SSLSocket = None) -> dict:
    if server_socket == None: server_socket = default_server_socket

    # starting the task
    task_list: list = [task_information.login_token.task_number] + task_parameters
    send_task(server_socket=server_socket, task=task_list)

    if task_parameters == task_information.login_token.login_task_parameters:
        information_dict: dict = {"username": username, "token": token}
        send_msg_as_json(information_dict, server_socket=server_socket)
    elif task_parameters == task_information.login_token.get_new_token_task_parameters:
        # returns "result", "token", "start_point" and "stop_point"
        information_dict: dict = {"old_token": old_token}
        send_msg_as_json(information_dict, server_socket=server_socket)
    elif task_parameters == task_information.login_token.delete_token_task_parameters:
        information_dict: dict = {"token": token}
        send_msg_as_json(information_dict, server_socket=server_socket)
    elif task_parameters == task_information.login_token.delete_all_tokens_task_parameters:
        pass # nothing has to be done here

    # return the outcome of the task
    task_result: dict = recv_msg_as_json(server_socket=server_socket)
    return task_result

def start_task_rsync(device: str, mountpoint: str, task_parameters: list = task_information.rsync.default_task_parameters, server_socket: ssl.SSLSocket = None):
    if server_socket == None: server_socket = default_server_socket

    # starting the task
    task_list: list = [task_information.rsync.task_number] + task_parameters
    send_task(server_socket=server_socket, task=task_list)

    information_dict = {"device": device, "mountpoint": mountpoint}
    send_msg_as_json(information_dict, server_socket=server_socket)

    result_json: dict = recv_msg_as_json(server_socket=server_socket)
    return result_json